/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1992 Electrotechnical Laboratry (ETL)

Permission to use, copy, modify, and distribute this material 
for any purpose and without fee is hereby granted, provided 
that the above copyright notice and this permission notice 
appear in all copies, and that the name of ETL not be 
used in advertising or publicity pertaining to this 
material without the specific, prior written permission 
of an authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY 
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", 
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=ISO-2022-JP
Program:	nnta.c (Network News Traffic Analizer)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
	 	Network News Traffic Analizer
History:
  v1.0 92-07-10  created
  v1.1    07-18  added -h M+N (thanks <yas@softlab.is.tsukuba.ac.jp>)
  v1.2    07-28  added histgram of delay
  v1.3    08-29  merged into nhc (news header cache)

TODO:
 - Path$@$N%[%C%W?t$K$h$k9J$j9~$_(J
 - Path$@$4$H$NCY1d$NJ,@O(J
 - $@F10l%=!<%9$4$H$N!"E~Ce4V3V(J($@J,I[(J)
 - $@5-;v%5%$%:$K4X$9$kE}7W(J
 - NNTP$@7PM3$N%R%9%H%j2r@O(J(HEAD,XHIST)
 - $@CY1d;~4V$NJ,I[$N%0%i%U(J($@J8;zC<KvMQ!"(Jxgraph$@MQ$J$I(J)
 - $@=PNO%U%)!<%^%C%H$N;XDj(J(%N=newsgroup %D=delay %P=path$@$J$I(J)
 - $@%X%C%@%-%c%C%7%e%i%$%V%i%j$H$N%j%s%/(J

///////////////////////////////////////////////////////////////////////*/

char *nnta_version = "1.3 92/08/29";

#include <stdio.h>
#include <time.h>
#include <strings.h>
char *strstr();
char *getenv();
double log2();

typedef unsigned int uint;

#define MAX_LINELEN	512
typedef char MessageLine[512];

#define DAYSEC		(24*60*60)

static int num_lines	= 0;
static int num_selected = 0;
static int num_filtered = 0;
static int num_canceled = 0;

static int num_opened	= 0;
static int num_headsizes[0x4000] = {0};
static int num_artsizes[64] = {0};
static int cum_headsize = 0;
static int cum_artsize	= 0;

static char *arg_historyfile	= HISTORY_FILE;	/* NEWS_HISTORY env. ver. */
static FILE *arg_historyfp	= NULL;
static char *arg_spooldir	= SPOOLDIR;	/* NEWS_SPOOL env. ver. */
static char *arg_pathfilter	= "";		/* -p			*/
static int   arg_pathskip	= 1;		/* -h N			*/
static int   arg_pathhops	= 1;		/* -h +M		*/
static uint  arg_begindate	= 0;		/* -s			*/
static uint  arg_enddate	= 0xFFFFFFFF;	/* -s			*/
static int   arg_printall	= 0;		/* -a			*/
static int   arg_min_delay	= 60;		/* -g (60 seconds)	*/
static int   arg_max_delay	= DAYSEC*7;	/* -l (7 days)		*/
static int   arg_verbose	= 0;		/* -v			*/
static int   arg_max_count	= 0x7FFFFFFF;	/* -m			*/
static int   arg_dumponly	= 0;		/* -d			*/
static int   arg_histgram	= 0;		/* -H			*/
static int   arg_sizehistgram	= 0;		/* -S			*/

#define get_path1(path,sub)	get_subpath(path,arg_pathskip,arg_pathhops,sub)

main(ac,av)
	char *av[];
{	FILE *hist;
	int lines;
	MessageLine line;
	char *ev;

	if( ev = getenv("NEWS_SPOOL") )
		arg_spooldir = ev;

	if( ev = getenv("NEWS_HISTORY") )
		arg_historyfile = ev;

	scan_args(ac,av);
	if( arg_historyfp )
		hist = arg_historyfp;
	else{
		hist = fopen(arg_historyfile,"r");
		if( hist == NULL ){
			fprintf(stderr,"%s: cannot open history-file '%s'\n",
				av[0],arg_historyfile);
			fprintf(stderr,"# setenv NEWS_HISTORY properly.\n");
			exit(1);
		}
	}
	if( access(arg_spooldir,0) != 0 ){
		fprintf(stderr,"%s: cannot access to '%s'\n",arg_spooldir);
		fprintf(stderr,"# setenv NEWS_SPOOL properly.\n");
		exit(2);
	}

	if( arg_begindate )
		fseek_history_bydate(hist,arg_begindate);

	for(num_lines = 1; fgets(line,sizeof(line),hist) != NULL; num_lines++){
		int rcode;

		rcode = 0;
		if( re_exec(line) == 1 ){
			num_filtered++;
			rcode = print_history1(line);
		}
		if( arg_verbose )
			print_verbose();
		if( arg_max_count < num_opened )
			break;
		if( rcode == -1 )
			break;
	}

	report_delay(0,0,""); /* flush */

	if( arg_sizehistgram )
		print_log2_histgram(num_artsizes,"S","bytes");
	fclose(hist);
}
static print_verbose(){
	int hi,cum;

	fprintf(stderr,"\r%d/%d/%d-%d ",
		num_selected,num_filtered,num_lines,num_canceled);

	cum = 0;
	for(hi = 0; hi < 512; hi++)
		cum += num_headsizes[hi];

	if( num_opened )
	fprintf(stderr,"H=%d/A=%d  H<512=%d%% ",
		cum_headsize/num_opened,cum_artsize/num_opened,
		cum*100 / num_opened);

	fflush(stderr);
}

static scan_args(ac,av)
	char *av[];
{	int ai;
	char *arg;

	re_comp(".");
	for(ai = 1; ai < ac; ai++){
		arg = av[ai];
		if( strcmp(arg,"-d") == 0 ){
			arg_dumponly = 1;
		}else
		if( strcmp(arg,"-a") == 0 ){
			arg_printall = 1;
		}else
		if( strcmp(arg,"-v") == 0 ){
			arg_verbose = 1;
		}else
		if( strcmp(arg,"-H") == 0 ){
			arg_histgram = 1;
		}else
		if( strcmp(arg,"-S") == 0 ){
			arg_sizehistgram = 1;
		}else
		if( strcmp(arg,"-") == 0 ){
			arg_historyfp = stdin;
		}else
		if( strcmp(arg,"-g") == 0 ){
			ai++;
			if( ai < ac ) arg_min_delay = atoi(av[ai]);
		}else
		if( strcmp(arg,"-l") == 0 ){
			ai++;
			if( ai < ac ) arg_max_delay = atoi(av[ai]) * DAYSEC;
		}else
		if( strcmp(arg,"-h") == 0 ){
			ai++;
			if( ai < ac )
				sscanf(av[ai],"%d+%d",
					&arg_pathskip,&arg_pathhops);
		}else
		if( strcmp(arg,"-p") == 0 ){
			ai++;
			if( ai < ac ) arg_pathfilter = av[ai];
		}else
		if( strcmp(arg,"-s") == 0 ){
			ai++;
			if( ai < ac ){
				char begin[128],*end;
				strcpy(begin,av[ai]);
				if( end = index(begin,',') )
					*end++ = 0;
				else	end = &begin[strlen(begin)];

				if( *begin ) arg_begindate = yymmddtime(begin);
				if( *end   ) arg_enddate = yymmddtime(end);
			}
		}else
		if( strcmp(arg,"-m") == 0 ){
			ai++;
			if( ai < ac ) arg_max_count = atoi(av[ai]);
		}else{
			re_comp(arg);
		}
	}
}

/***********************************************************************
 *	Path accumulater
 */
#define NUM_SOURCES	64

typedef struct {
	char	pf_path[64];
	int	pf_hops;
	int	pf_count;
} PathFreq;

typedef struct {
	int	tf_count;
	double	tf_delays;
	int	tf_delayhist[64];
	int	tf_xcount;
     struct tm	tf_tm;
     PathFreq	tf_pathfreq[NUM_SOURCES];
} Traffic;

static Traffic day_traffic;
static Traffic week_traffic;
static Traffic month_traffic;
static Traffic itraffic;

static int TZOFFSET = 9*60*60;
static report_delay(pdate,adate,path)
	char *path;
{	struct tm tm;

	tm = *localtime(&adate);
	if( !adate || tm.tm_mday != day_traffic.tf_tm.tm_mday )
		print_traffic(&day_traffic,"d ");
	if( !adate || get_week(&tm) != get_week(&week_traffic.tf_tm) )
		print_traffic(&week_traffic,"w ");
	if( !adate || tm.tm_mon != month_traffic.tf_tm.tm_mon )
		print_traffic(&month_traffic,"m ");

	day_traffic.tf_tm = tm;
	week_traffic.tf_tm = tm;
	month_traffic.tf_tm = tm;

	if( adate && pdate ){
		int delay;

		delay = adate - pdate;
		count_traffic(&month_traffic,path,delay);
		count_traffic(&week_traffic,path,delay);
		count_traffic(&day_traffic,path,delay);
	}
}

static print_traffic(traffic,prefix)
	Traffic *traffic;
	char *prefix;
{	int adate,day,mdelay,meanhops;
	char sdelay[256],pathes[4096];

	if(traffic->tf_xcount){
		adate = timelocal(&traffic->tf_tm);
		if( traffic->tf_count )
			mdelay = traffic->tf_delays/traffic->tf_count;
		else	mdelay = 0;
		sprint_delay(sdelay,mdelay);
		meanhops = sprint_pathes(pathes,traffic->tf_pathfreq);

		printf("%s",prefix);
		print_date(stdout,adate,0);
		printf(" %s %5d*%d-%-3d ",sdelay,
			traffic->tf_xcount,meanhops,
			traffic->tf_xcount-traffic->tf_count);
		printf("%s\n",pathes);

		if( arg_histgram )
			print_log2_histgram(traffic->tf_delayhist,"H","min");
		*traffic = itraffic;
	}
}
static print_log2_histgram(iarray,prefix,unit)
	int *iarray,*prefix,*unit;
{	int hi,minhi,maxhi,count,total,cum,percent,cumpercent,pi;

	total = 0;
	minhi = -1;
	maxhi = -1;
	for(hi = 0; hi < 64; hi++){
		if( count = iarray[hi] ){
			if( minhi == -1 )
				minhi = hi;
			total += count;
			maxhi = hi;
		}
	}
	if( maxhi == -1 )
		return;
	cum = 0;
	for(hi = minhi; hi <= maxhi; hi++){
		count = iarray[hi];
		cum += count;
		cumpercent = cum*100/total;
		printf("%s %02d ",prefix,hi-32);
		if( 32 <= hi )
			printf("%5d",(1<<(hi-32)));
		else	printf("%5.3f",1/(double)(1<<(32-hi)));
		printf("%s-%5d %3d%%  ",unit,count,cumpercent);

		percent = count*100/total;
		for(pi = 0; pi < percent/2; pi++)
			printf("*");
		for(; pi < cumpercent/2; pi++)
			printf("|");
		printf("|\n");
	}
}
static count_traffic(traffic,path,delay)
	Traffic *traffic;
	char *path;
{	int logt;

	if( arg_min_delay < delay && delay < arg_max_delay ){
		traffic->tf_count++;
		traffic->tf_delays += delay;
		logt = (int)log2((double)delay/60)+32;
		traffic->tf_delayhist[logt] += 1;
	}
	traffic->tf_xcount++;
	count_path(traffic->tf_pathfreq,path);
}
static count_path(pathfreq,path)
	PathFreq *pathfreq;
	char *path;
{	int pi,hops; char *xp;
	MessageLine path1;

	hops = get_path1(path,path1);
	for(pi = 0; *(xp = pathfreq[pi].pf_path); pi++)
		if(strcmp(xp,path1) == 0)
			break;
	strcpy(pathfreq[pi].pf_path,path1);
	pathfreq[pi].pf_count += 1;
	pathfreq[pi].pf_hops += hops;
}

static sort_pathfreq(a,b)
	PathFreq *a,*b;
{
	return b->pf_count - a->pf_count;
}
static sprint_pathes(pathes,pathfreq)
	char *pathes;
	PathFreq *pathfreq;
{	int pi,pn,thops,tarts;
	PathFreq counts[NUM_SOURCES];
	char *pp;

	for(pi = 0; pi < NUM_SOURCES; pi++){
		counts[pi] = pathfreq[pi];
		if( counts[pi].pf_count == 0 )
			break;
	}
	pn = pi;

	if( 1 < pn )
	qsort(counts,pn,sizeof(PathFreq),sort_pathfreq);

	thops = tarts = 0;
	pp = pathes;
	*pp = 0;
	for(pi = 0; pi < pn; pi++){
		thops += counts[pi].pf_hops;
		tarts += counts[pi].pf_count;

		sprintf(pp,"%s(%d*%d) ",
			counts[pi].pf_path,
			counts[pi].pf_count,
			counts[pi].pf_hops/counts[pi].pf_count);
		pp += strlen(pp);
	}
	if( tarts )
		return thops/tarts;
	else	return 0;
}

/***********************************************************************
 *	History and header parser
 */
static FILE* fopen_article(group,number)
	char *group;
{	char path[1024],*p;

	sprintf(path,"%s/%s/%d",arg_spooldir,group,number);
/*
	sprintf(path,"%s/%d",group,number);
*/
	for(p = path; *p; p++)
		if( *p == '.' )
			*p = '/';
	return fopen(path,"r");
}

static print_history1(line)
	char *line;
{	MessageLine mid,xref,adates,posted,path,path1;
	uint adate,pdate;
	char group[256];
	int number;
	FILE *afp;

	adate = decomp_history_line(line,mid,xref);
	if( adate < arg_begindate )
		return 0;
	if( adate > arg_enddate)
		return -1;
	if( arg_dumponly ){
		printf("%d\t%s\t%s\n",adate,mid,xref);
		return 0;
	}
	if( sscanf(xref,"%[^/]/%d",group,&number) != 2 ){
		num_canceled++;
		return 0;
	}

	if((afp = fopen_article(group,number)) == NULL)
		num_canceled++;
	else{
		int hsize,asize;

		num_opened++;
		asize = get_artsize(afp,&hsize);
		cum_artsize += asize;
		num_artsizes[(int)log2((double)asize)+32] += 1;
		cum_headsize += hsize;
		num_headsizes[hsize] += 1;

		get_field(afp,"Date",posted);
		pdate = parse_date(posted);

		get_field(afp,"Path",path);
		get_path1(path,path1);

		if( strstr(path,arg_pathfilter) ){
			num_selected++;
			report_delay(pdate,adate,path);
			if( arg_printall )
				print_article(afp,group,number,
					pdate,adate,path,xref);
		}
		fclose(afp);
	}
	return 0;
}

static print_article(afp,group,number,pdate,adate,path,xref)
	FILE *afp;
	char *group,*path,*xref;
{	MessageLine from,path1;

	get_field(afp,"From",from);

	printf("a ");
	print_date(stdout,adate,1);
	if( adate && pdate && 0 <= (adate-pdate) ){
		char sdelay[32];
		sprint_delay(sdelay,adate-pdate);
		printf(" %10s ",sdelay);
	}else	printf(" %10s ","?");
	get_path1(path,path1);
	printf("%-14s",path1);

	if( 0 ){
		printf(" ");
		print_from(stdout,from);
	}else{
		if( *xref )
			printf(" %s",xref);
		else	printf(" %s/%d",group,number);
	}
	printf("\n");
}
static print_from(fp,from)
	FILE *fp;
	char *from;
{	MessageLine addr;

	*addr = 0;
	sscanf(from,"%s",addr);
	fprintf(fp,"%s",addr);
}
static get_field(afp,tag,field)
	FILE *afp;
	char *tag,*field;
{	MessageLine line;
	int len = strlen(tag);

	fseek(afp,0,0);
	*field = 0;
	while( fgets(line,sizeof(line),afp) != NULL ){
		if( *line == '\n' )
			break;
		if( !strncmp(tag,line,len) && !strncmp(": ",line+len,2) ){
			strcpy(field,line+len+2);
			return 1;
		}
	}
	return 0;
}
static get_artsize(afp,hsizep)
	FILE *afp;
	int *hsizep;
{	MessageLine line;
	int asize = 0, hsize = 0;

	fseek(afp,0,0);
	while( fgets(line,sizeof(line),afp) != NULL ){
		hsize += strlen(line);
		if( strcmp(line,"\n") == 0 || strcmp(line,"\r\n") == 0 )
			break;
	}
	asize = hsize;
	while( fgets(line,sizeof(line),afp) != NULL )
		asize += strlen(line);
	if( hsizep )
		*hsizep = hsize;
	return asize;
}
static get_subpath(path,skip,hops,subpath)
	char *path,*subpath;
{	int nodes[256],pi,pc,nnodes = 0;

	if( *path == 0 )
		return 0;

	nodes[nnodes++] = 0;
	for(pi = 0; pc = path[pi]; pi++)
		if(pc == '!')
			nodes[nnodes++] = pi+1;

	if( nnodes <= 2 ){ strcpy(subpath,"LOCAL"); return 0; }

	if( nnodes < skip ){
		*subpath = 0;
		return 0;
	}else{
		strcpy(subpath,&path[nodes[skip]]);
		if(skip+hops < nnodes)
			subpath[nodes[skip+hops]-nodes[skip]-1] = 0;
		return nnodes;
	}
}

/***********************************************************************
 *	Date parser/printer
 */
static sprint_delay(delay,clock)
	char *delay;
{	int hour,min,sec;

	hour =  clock / (60*60);
	min  = (clock / 60) % 60;
	sec  =  clock % 60;
	sprintf(delay,"%3d:%02d:%02d",hour,min,sec);
}

static struct tm zero_tm;
static parse_date(sdate)
	char *sdate;
{	char mon[32],zone[32],*comma;
	int inum,date,offs,offclock;
	struct tm tm;

	tm  = zero_tm;
	if( comma = (char*)index(sdate,',') )
		sdate = comma + 1;

	inum = sscanf(sdate," %d %s %d %d:%d:%d %s",
		&tm.tm_mday, mon, &tm.tm_year,
		&tm.tm_hour, &tm.tm_min, &tm.tm_sec,
		zone
	);
	if( inum != 7 )
		return 0;

	if( 1900 < tm.tm_year )
		tm.tm_year -= 1900;
	tm.tm_mon = montoi(mon);

	if( strcmp(zone,"GMT") != 0 ){
		offs = atoi(zone);
		if( offs < -1200 || 1200 < offs ){
			fprintf(stderr,"??? %s\n",sdate);
			return 0;
		}
		offclock = (offs/100)*60*60 + (offs%100)*60;
	}else	offclock = 0;

	return timegm(&tm) + offclock;
}
static yymmddtime(yymmdd)
	char *yymmdd;
{	struct tm tm;

	tm = zero_tm;
	tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
	if(sscanf(yymmdd,"%d/%d/%d+%d:%d:%d",
		&tm.tm_year, &tm.tm_mon, &tm.tm_mday,
		&tm.tm_hour, &tm.tm_min, &tm.tm_sec) < 3)
			return 0;
	tm.tm_mon--;
	return timelocal(&tm);
}

static char *months[]=
{"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec",0};
static montoi(mon)
	char *mon;
{	int mi;

	for(mi = 0; mi < 12; mi++)
		if(strcmp(months[mi],mon) == 0)
			return mi;
	return 0;
}
static char *weeks[] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat",""};
static print_date(fp,clock,timetoo)
	FILE *fp;
{	struct tm *tm;

	tm = localtime(&clock);
	fprintf(fp,"%s ",weeks[tm->tm_wday]);
	fprintf(fp, "%02d/%02d/%02d",tm->tm_year, tm->tm_mon+1, tm->tm_mday);
	if( timetoo )
	fprintf(fp," %02d:%02d:%02d",tm->tm_hour, tm->tm_min,   tm->tm_sec);

}
static get_week(tm)
	struct tm *tm;
{	int clock;

	clock = timegm(tm);
	clock /= 60*60*24;
	clock -= tm->tm_wday;
	return clock;
}

/*//////////////////////////////////////////////////////////////////////
 *	MISCELLANEOUS
 */
static char *strstr(s1,s2)
	char *s1,*s2;
{	char *p1;
	int len;

	if( *s2 == 0 )
		return s1;

	len = strlen(s2);
	for( p1 = s1; *p1; p1 ++ )
		if( *p1 == *s2 && strncmp(p1,s2,len)==0 ) 
			return p1;
	return 0;
}
